from typing import List, Iterable, Dict, Tuple, Any


class LanguageTokenizer:

    """Simple tokenizer for language to index mapping."""

    def __init__(self, languages: List[str]) -> None:
        """
        Initializes a language tokenizer for a list of languages.

        Args:
          languages (List[str]): List of languages, e.g. ['de', 'en'].
        """

        self.lang_index = {l: i for i, l in enumerate(languages)}
        self.index_lang = {i: l for i, l in enumerate(languages)}

    def __call__(self, lang: str) -> int:
        """
        Maps the language to an index.

        Args:
          lang (str): Language to be mapped, e.g. 'de'.

        Returns:
          int: Index of language.
        """

        if lang not in self.lang_index:
            raise ValueError(f'Language not supported: {lang}. '
                             f'Supported languages: {self.lang_index.keys()}')

        return self.lang_index[lang]

    def decode(self, index: int) -> str:
        """Inverts the index mapping of a language.

        Args:
          index (int): Index of language.

        Returns:
          str: Language for the given index.

        """
        return self.index_lang[index]


class SequenceTokenizer:

    """Tokenizes text and optionally attaches language-specific start index (and non-specific end index)."""

    def __init__(self,
                 symbols: List[str],
                 languages: List[str],
                 char_repeats: int,
                 lowercase: bool = True,
                 append_start_end: bool =True,
                 pad_token='_',
                 end_token='<end>') -> None:
        """
        Initializes a SequenceTokenizer object.

        Args:
          symbols (List[str]): Character (or phoneme) symbols.
          languages (List[str]): List of languages.
          char_repeats (int): Number of repeats for each character to allow the forward model to map to longer
               phoneme sequences. Example: for char_repeats=2 the tokenizer maps hi -> hhii.
          lowercase (bool): Whether to lowercase the input word.
          append_start_end (bool): Whether to append special start and end tokens. Start and end tokens are
               index mappings of the chosen language.
          pad_token (str): Special pad token for index 0.
          end_token (str): Special end of sequence token.
        """

        self.languages = languages
        self.lowercase = lowercase
        self.char_repeats = char_repeats
        self.append_start_end = append_start_end
        self.pad_index = 0
        self.token_to_idx = {pad_token: self.pad_index}
        self.special_tokens = {pad_token, end_token}
        for lang in languages:
            lang_token = self._make_start_token(lang)
            self.token_to_idx[lang_token] = len(self.token_to_idx)
            self.special_tokens.add(lang_token)
        self.token_to_idx[end_token] = len(self.token_to_idx)
        self.end_index = self.token_to_idx[end_token]
        for symbol in symbols:
            self.token_to_idx[symbol] = len(self.token_to_idx)
        self.idx_to_token = {i: s for s, i in self.token_to_idx.items()}
        self.vocab_size = len(self.idx_to_token)

    def __call__(self, sentence: Iterable[str], language: str) -> List[int]:
        """
        Maps a sequence of symbols for a language to a sequence of indices.

        Args:
          sentence  (Iterable[str]): Sentence (or word) as a sequence of symbols.
          language (str): Language for the mapping that defines the start and end token indices.

        Returns:
           List[int]: Sequence of token indices.
        """

        sentence = [item for item in sentence for i in range(self.char_repeats)]
        if language not in self.languages:
            raise ValueError(f'Language not supported: {language}. Supported languages: {self.languages}')
        if self.lowercase:
            sentence = [s.lower() for s in sentence]
        sequence = [self.token_to_idx[c] for c in sentence if c in self.token_to_idx]
        if self.append_start_end:
            sequence = [self._get_start_index(language)] + sequence + [self.end_index]
        return sequence

    def decode(self, sequence: Iterable[int], remove_special_tokens: bool = False) -> List[str]:
        """Maps a sequence of indices to a sequence of symbols.

        Args:
          sequence (Iterable[int]): Encoded sequence to be decoded.
          remove_special_tokens (bool): Whether to remove special tokens such as pad or start and end tokens. (Default value = False)
          sequence: Iterable[int]: 

        Returns:
           List[str]: Decoded sequence of symbols.
        """

        sequence = list(sequence)
        if self.append_start_end:
            sequence = sequence[:1] + sequence[1:-1:self.char_repeats] + sequence[-1:]
        else:
            sequence = sequence[::self.char_repeats]
        decoded = [self.idx_to_token[int(t)] for t in sequence if int(t) in self.idx_to_token]
        if remove_special_tokens:
            decoded = self.remove_special_tokens(decoded)
        return decoded

    def remove_special_tokens(self, decoded: List[str]) -> List[str]:
        """Removes special tokens from a decoded sequence.

        Args:
          decoded (List[str]): Decoded sequence of symbols.

        Returns:
          List[str]: Decoded sequence with special tokens removed.
        """
        return [d for d in decoded if d not in self.special_tokens]

    def _get_start_index(self, language: str) -> int:
        lang_token = self._make_start_token(language)
        return self.token_to_idx[lang_token]

    def _make_start_token(self, language: str) -> str:
        return '<' + language + '>'


class Preprocessor:

    """ Preprocesses data for a phonemizer training session. """

    def __init__(self,
                 lang_tokenizer: LanguageTokenizer,
                 text_tokenizer: SequenceTokenizer,
                 phoneme_tokenizer: SequenceTokenizer) -> None:
        """
        Initializes a preprocessor object.

        Args:
          lang_tokenizer (LanguageTokenizer): Tokenizer for input language.
          text_tokenizer (SequenceTokenizer): Tokenizer for input text.
          phoneme_tokenizer (SequenceTokenizer): Tokenizer for output phonemes.
        """

        self.lang_tokenizer = lang_tokenizer
        self.text_tokenizer = text_tokenizer
        self.phoneme_tokenizer = phoneme_tokenizer

    def __call__(self,
                 item: Tuple[str, Iterable[str], Iterable[str]]) \
            -> Tuple[int, List[int], List[int]]:
        """
        Preprocesses a data point.

        Args:
          item (Tuple): Data point comprised of (language, input text, output phonemes).

        Returns: Tuple: Preprocessed data point as (language tokens, input_text tokens, output phonemes tokens)
        """

        lang, text, phonemes = item
        lang_token = self.lang_tokenizer(lang)
        text_tokens = self.text_tokenizer(text, lang)
        phoneme_tokens = self.phoneme_tokenizer(phonemes, lang)
        return lang_token, text_tokens, phoneme_tokens

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> 'Preprocessor':
        """Initializes a preprocessor from a config.

        Args:
          config (Dict[str, Any]): Dictionary containing preprocessing hyperparams.

        Returns:
          Preprocessor: Preprocessor object.
        """

        text_symbols = config['preprocessing']['text_symbols']
        phoneme_symbols = config['preprocessing']['phoneme_symbols']
        lang_symbols = config['preprocessing']['languages']
        char_repeats = config['preprocessing']['char_repeats']
        lowercase = config['preprocessing']['lowercase']
        lang_tokenizer = LanguageTokenizer(lang_symbols)
        text_tokenizer = SequenceTokenizer(symbols=text_symbols,
                                           languages=lang_symbols,
                                           char_repeats=char_repeats,
                                           lowercase=lowercase,
                                           append_start_end=True)
        phoneme_tokenizer = SequenceTokenizer(phoneme_symbols,
                                              languages=lang_symbols,
                                              lowercase=False,
                                              char_repeats=1,
                                              append_start_end=True)
        return Preprocessor(lang_tokenizer=lang_tokenizer,
                            text_tokenizer=text_tokenizer,
                            phoneme_tokenizer=phoneme_tokenizer)



import unittest

from dp.preprocessing.text import LanguageTokenizer


class TestLanguageTokenizer(unittest.TestCase):

    def test_call_happy_path(self) -> None:

        tokenizer = LanguageTokenizer(languages=['de', 'en'])

        de_tok = tokenizer('de')
        en_tok = tokenizer('en')
        self.assertEqual(0, de_tok)
        self.assertEqual(1, en_tok)

        self.assertEqual('de', tokenizer.decode(de_tok))
        self.assertEqual('en', tokenizer.decode(en_tok))

    def test_exception(self) -> None:
        tokenizer = LanguageTokenizer(languages=['de', 'en'])

        with self.assertRaises(ValueError):
            tokenizer('not_existent')




import os
import shutil
import tempfile
import unittest
from pathlib import Path

from dp import preprocess
from dp.preprocess import preprocess
from dp.preprocessing.text import Preprocessor
from dp.utils.io import read_config, unpickle_binary, save_config


class TestPreprocess(unittest.TestCase):

    def setUp(self) -> None:
        temp_dir = tempfile.mkdtemp(prefix='TestPreprocessTmp')
        self.temp_dir = Path(temp_dir)
        test_path = os.path.dirname(os.path.abspath(__file__))
        self.test_config_path = Path(test_path) / 'resources/forward_test_config.yaml'

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_dir, ignore_errors=True)

    def test_preprocess_happy_path(self) -> None:
        config = read_config(self.test_config_path)
        data_dir = self.temp_dir / 'datasets'
        config_path = self.temp_dir / 'forward_test_config.yaml'
        config['paths']['data_dir'] = str(data_dir)
        save_config(config, config_path)

        train_data = [
            ('de','benützten', 'bəˈnʏt͡stn̩'),
            ('de', 'gewürz', 'ɡəˈvʏʁt͡s'),
            ('en_us', 'young', 'jʌŋ'),
        ]

        val_data = [
            ('en_us', 'young', 'jʌŋ'),
            ('de', 'gewürz', 'ɡəˈvʏʁt͡s')
        ]

        preprocess(config_file=config_path,
                   train_data=train_data,
                   val_data=val_data,
                   deduplicate_train_data=False)

        preprocessor = Preprocessor.from_config(config)

        expected_train = [preprocessor(t) for t in train_data]
        actual_train = unpickle_binary(data_dir / 'train_dataset.pkl')
        self.assertEqual(expected_train, actual_train)

        expected_val = [preprocessor(t) for t in val_data]
        actual_val = unpickle_binary(data_dir / 'val_dataset.pkl')
        self.assertEqual(expected_val, actual_val)

        expected_phon_dict = {'de': {}, 'en_us': {}}
        for lang, word, phon in train_data + val_data:
            expected_phon_dict[lang][word] = phon

        actual_phon_dict = unpickle_binary(data_dir / 'phoneme_dict.pkl')
        self.assertEqual(expected_phon_dict, actual_phon_dict)


import unittest
from typing import Dict, Any, Tuple
from unittest.mock import patch

import torch

from dp.model.model import Model
from dp.model.predictor import Predictor
from dp.preprocessing.text import Preprocessor


class ModelMock:

    def generate(self, batch: Dict[str, Any], **kwargs) -> Tuple[torch.tensor, torch.tensor]:
        """ Return input and ones as probs """
        tokens = batch['text']
        probs = torch.ones(tokens.size())
        return tokens, probs


@patch.object(Model, 'generate', new_callable=ModelMock)
class TestPredictor(unittest.TestCase):

    def test_call_with_model_mock(self, model_mock: Model) -> None:
        config = {
            'preprocessing': {
                'text_symbols': 'abcd',
                'phoneme_symbols': 'abcd',
                'char_repeats': 1,
                'languages': ['de'],
                'lowercase': False
            },
        }
        preprocessor = Preprocessor.from_config(config)
        predictor = Predictor(model_mock, preprocessor)
        texts = ['ab', 'cde']

        result = predictor(texts, lang='de', batch_size=8)
        self.assertEqual(2, len(result))
        self.assertEqual('ab', result[0].word)
        self.assertEqual('ab', result[0].phonemes)
        self.assertEqual('cd', result[1].phonemes)

        result = predictor(texts, lang='de', batch_size=1)
        self.assertEqual(2, len(result))
        self.assertEqual('ab', result[0].phonemes)
        self.assertEqual('cd',result[1]. phonemes)

        texts = ['/']
        result = predictor(texts, lang='de', batch_size=1)
        self.assertEqual(1, len(result))
        self.assertEqual('', result[0].phonemes)
        self.assertEqual([], result[0].phoneme_tokens)



import unittest

from dp.preprocessing.text import Preprocessor


class TestPreprocessor(unittest.TestCase):

    def test_call_happy_path(self) -> None:
        config = {
            'preprocessing': {
                'text_symbols': 'abcdABCD',
                'phoneme_symbols': 'abcd',
                'char_repeats': 1,
                'languages': ['de', 'en'],
                'lowercase': False
            },
        }

        preprocessor = Preprocessor.from_config(config)
        result = preprocessor(('de', 'aB', 'cd'))
        lang, text, phons = result
        lang = preprocessor.lang_tokenizer.decode(lang)
        text = preprocessor.text_tokenizer.decode(text, remove_special_tokens=False)
        phons = preprocessor.phoneme_tokenizer.decode(phons, remove_special_tokens=False)
        self.assertEqual('de', lang)
        self.assertEqual(['<de>', 'a', 'B', '<end>'], text)
        self.assertEqual(['<de>', 'c', 'd', '<end>'], phons)


from collections import Counter
from pathlib import Path
from random import Random
from typing import List, Tuple, Iterable

import tqdm

from dp.model.model import ModelType
from dp.preprocessing.text import Preprocessor
from dp.utils.io import read_config, pickle_binary
from dp.utils.logging import get_logger

logger = get_logger(__name__)


def preprocess(config_file: str,
               train_data: List[Tuple[str, Iterable[str], Iterable[str]]],
               val_data: List[Tuple[str, Iterable[str], Iterable[str]]] = None,
               deduplicate_train_data=True) -> None:
    """
    Preprocesses a given dataset to enable model training. The preprocessing result is stored in
    a folder provied by the config.

    Args:
      config_file (str): Path to the config.yaml that provides all necessary hyperparameters.
      train_data (List[Tuple[str, Iterable[str], Iterable[str]]]):
        Training data as a list of Tuples (language, grapheme sequence, phoneme sequence).
      val_data (List[Tuple[str, Iterable[str], Iterable[str]]], optional):
        Validation data as a list of Tuples (language, grapheme sequence, phoneme sequence).
      deduplicate_train_data (bool): Whether to deduplicate multiple occurences of the same word,
                                    the first is taken (Default value = True).

    Returns:
      None: the preprocessing result is stored in a folder provided by the config.

    """

    config = read_config(config_file)

    model_type = config['model']['type']
    model_type = ModelType(model_type)
    if model_type.is_autoregressive() and config['preprocessing']['char_repeats'] > 1:
        char_repeats = config['preprocessing']['char_repeats']
        logger.warning(f'WARNING: You are training autoregressive model with char_repeats={char_repeats}. '
                       f'It is recommended to set char_repeats=1 in the config and preprocess again.')

    languages = set(config['preprocessing']['languages'])

    logger.info(f'Preprocessing, train data: with {len(train_data)} files.')

    data_dir = Path(config['paths']['data_dir'])
    data_dir.mkdir(parents=True, exist_ok=True)

    train_dict = {(l, w): [] for l, w, p in train_data}
    for l, w, p in train_data:
        train_dict[(l, w)] = train_dict[(l, w)] + [(l, w, p)]
    train_keys = sorted(list(train_dict.keys()))

    if val_data is not None:
        val_data = [(l, w, p) for l, w, p in val_data if l in languages]
    else:
        n_val = config['preprocessing']['n_val']
        logger.info(f'Performing random split with num val: {n_val}')
        random = Random(42)
        random.shuffle(train_keys)
        val_keys = train_keys[:n_val]
        train_keys = train_keys[n_val:]
        val_data = []
        for k in val_keys:
            val_data.extend(train_dict[k])

    train_data = []
    for key in train_keys:
        data_list = train_dict[key]
        if deduplicate_train_data:
            train_data.append(data_list[0])
        else:
            train_data.extend(data_list)

    preprocessor = Preprocessor.from_config(config)

    train_count = Counter()
    val_count = Counter()

    logger.info('Processing train data...')
    train_dataset = []
    for i, (lang, text, phonemes) in enumerate(tqdm.tqdm(train_data, total=len(train_data))):
        tokens = preprocessor((lang, text, phonemes))
        train_dataset.append(tokens)
        train_count.update([lang])

    val_dataset = []
    for i, (lang, text, phonemes) in enumerate(val_data):
        tokens = preprocessor((lang, text, phonemes))
        val_dataset.append(tokens)
        val_count.update([lang])

    logger.info(f'\nSaving datasets to: {data_dir.absolute()}')
    pickle_binary(train_dataset, data_dir / 'train_dataset.pkl')
    pickle_binary(val_dataset, data_dir / 'val_dataset.pkl')
    phoneme_dictionary = dict()

    all_data = []
    text_symbols = set(config['preprocessing']['text_symbols'])
    phoneme_symbols = set(config['preprocessing']['phoneme_symbols'])
    for lang, text, phon in sorted(train_data + val_data):
        text = ''.join([t for t in text if t in text_symbols])
        phons = ''.join([p for p in phon if p in phoneme_symbols])
        all_data.append((lang, text, phons))

    for l, w, p in all_data:
        lang_dict = phoneme_dictionary.setdefault(l, {})
        if w not in lang_dict:
            lang_dict[w] = p

    pickle_binary(phoneme_dictionary, data_dir / 'phoneme_dict.pkl')
    with open(data_dir / 'combined_dataset.txt', 'w+', encoding='utf-8') as f:
        for lang, text, phoneme in all_data:
            f.write(f'{lang}\t{text}\t{phoneme}\n')

    logger.info(f'Preprocessing. \nTrain counts (deduplicated): {train_count.most_common()}'
                f'\nVal counts (including duplicates): {val_count.most_common()}')

    assert len(train_count) > 0, 'Preprocessing resulted in zero train counts!'
    assert len(val_count) > 0, 'Preprocessing resulted in zero validation counts!'



import os
from pathlib import Path

import torch
from torch.distributed import init_process_group

from dp.model.model import load_checkpoint, ModelType, \
    create_model
from dp.preprocessing.text import Preprocessor
from dp.training.trainer import Trainer
from dp.utils.io import read_config
from dp.utils.logging import get_logger

logger = get_logger(__name__)


def train(rank: int,
          num_gpus: int,
          config_file: str,
          checkpoint_file: str = None) -> None:
    """
    Runs training of a transformer model.

    Args:
      rank (int): Device id
      num_gpus (int): Number of devices
      config_file (str): Path to the config.yaml that stores all necessary parameters.
      checkpoint_file (str, optional): Path to a model checkpoint to resume training for (e.g. latest_model.pt)

    Returns:
        None: The model checkpoints are stored in a folder provided by the config.

    """

    config = read_config(config_file)

    if num_gpus > 1:
        os.environ["MASTER_ADDR"] = config['training']['ddp_host']
        os.environ["MASTER_PORT"] = config['training']['ddp_post']
        init_process_group(backend=config['training']['ddp_backend'], rank=rank, world_size=num_gpus)

    if checkpoint_file is not None:
        logger.info(f'Restoring model from checkpoint: {checkpoint_file}')
        model, checkpoint = load_checkpoint(checkpoint_file)
        model.train()
        step = checkpoint['step']
        logger.info(f'Loaded model with step: {step}')
        for key, val in config['training'].items():
            val_orig = checkpoint['config']['training'][key]
            if val_orig != val:
                logger.info(f'Overwriting training param: {key} {val_orig} --> {val}')
                checkpoint['config']['training'][key] = val
        config = checkpoint['config']
        model_type = config['model']['type']
        model_type = ModelType(model_type)
    else:
        logger.info('Initializing new model from config...')
        preprocessor = Preprocessor.from_config(config)
        model_type = config['model']['type']
        model_type = ModelType(model_type)
        model = create_model(model_type, config=config)
        checkpoint = {
            'preprocessor': preprocessor,
            'config': config,
        }

    checkpoint_dir = Path(config['paths']['checkpoint_dir'])
    logger.info(f'Checkpoints will be stored at {checkpoint_dir.absolute()}')
    loss_type = 'cross_entropy' if model_type.is_autoregressive() else 'ctc'

    if num_gpus > 0:
        device = torch.device(f'cuda:{rank}')
    else:
        device = torch.device('cpu')

    use_ddp = True if num_gpus > 1 else False

    trainer = Trainer(checkpoint_dir=checkpoint_dir, device=device, rank=rank, use_ddp=use_ddp, loss_type=loss_type)
    trainer.train(model=model,
                  checkpoint=checkpoint,
                  store_phoneme_dict_in_model=config['training']['store_phoneme_dict_in_model'])



import unittest
from typing import List
from unittest.mock import patch

from dp.phonemizer import Phonemizer
from dp.model.predictor import Predictor, Prediction


class PredictorMock:

    def __call__(self, words: List[str], lang: str, **kwargs) -> List[Prediction]:
        """ Simply returns the original words with a suffix -phon-{language} to test for calls. """

        output = []
        for word in words:
            phons = word + f'-phon-{lang}'
            pred = Prediction(word=word, phonemes=phons, confidence=0.5,
                              phoneme_tokens=list(phons), token_probs=[0.5]*len(phons))
            output.append(pred)
        return output


class TestPhonemizer(unittest.TestCase):

    @patch.object(Predictor, '__call__', new_callable=PredictorMock)
    def test_call_with_predictor_mock(self, predictor: Predictor) -> None:
        phonemizer = Phonemizer(predictor=predictor)

        result = phonemizer('hallo', lang='de')
        self.assertEqual('hallo-phon-de', result)

        result = phonemizer(['hallo', 'du'], lang='de')
        self.assertEqual(['hallo-phon-de', 'du-phon-de'], result)

    @patch.object(Predictor, '__call__', new_callable=PredictorMock)
    def test_apply_dictionary(self, predictor: Predictor) -> None:
        phoneme_dict = {'de': {}, 'en': {}}
        phoneme_dict['de']['die'] = 'DIE'
        phoneme_dict['de']['das'] = 'DAS'
        phoneme_dict['de']['das-die'] = 'DAS!DIE'
        phoneme_dict['en']['die'] = 'DAI'

        phonemizer = Phonemizer(predictor=predictor,
                                lang_phoneme_dict=phoneme_dict)

        result = phonemizer('die', lang='de')
        self.assertEqual('DIE', result)

        result = phonemizer('Die', lang='de')
        self.assertEqual('DIE', result)

        result = phonemizer('DIE', lang='de')
        self.assertEqual('DIE', result)

        # test whether the dict check gets hyphenated words
        result = phonemizer('das-die', lang='de')
        self.assertEqual('DAS!DIE', result)

        # test whether the dict check is applied to subwords in a hyphenated word
        result = phonemizer('die-das', lang='de')
        self.assertEqual('DIE-DAS', result)

        result = phonemizer('/die/?!', lang='de', punctuation='')
        self.assertEqual('DIE', result)

        result = phonemizer('/die/?!', lang='de', punctuation='/')
        self.assertEqual('/DIE/', result)

        result = phonemizer('die', lang='en')
        self.assertEqual('DAI', result)

        result = phonemizer('die-die-das', lang='en')
        self.assertEqual('DAI-DAI-das-phon-en', result)

        result = phonemizer('dies', lang='de')
        self.assertEqual('dies-phon-de', result)

        result = phonemizer('dies', lang='en')
        self.assertEqual('dies-phon-en', result)

    @patch.object(Predictor, '__call__', new_callable=PredictorMock)
    def test_split_result(self, predictor: Predictor) -> None:
        phonemizer = Phonemizer(predictor=predictor)

        text = 'Ich, bin - so wie (ich/du) bin.'
        result = phonemizer.phonemise_list([text], lang='de', punctuation=',().')
        cleaned_text = ''.join(result.split_text[0])
        self.assertEqual('Ich, bin - so wie (ichdu) bin.', cleaned_text)
        self.assertEqual(len(result.split_text[0]), len(result.split_phonemes[0]))

        text = 'Ich, bin: - (so.'
        result = phonemizer.phonemise_list([text], lang='de', punctuation=',().')
        self.assertEqual(['Ich', ',', ' ', 'bin', ' ', '-', ' ', '(', 'so', '.'],
                         result.split_text[0])

        # test some boundaries
        result = phonemizer('whitespace ', lang='de')
        self.assertEqual('whitespace-phon-de ', result)

        # test some boundaries
        result = phonemizer(',special!?', lang='de', punctuation=',!')
        self.assertEqual(',special-phon-de!', result)

        # test some boundaries
        result = phonemizer(',special!?', lang='de', punctuation='')
        self.assertEqual('special-phon-de', result)

    def test_expand_acronym(self) -> None:
        expanded = Phonemizer._expand_acronym('hallo')
        self.assertEqual('hallo', expanded)

        expanded = Phonemizer._expand_acronym('HAL')
        self.assertEqual('H-A-L', expanded)

        expanded = Phonemizer._expand_acronym('hallO')
        self.assertEqual('hall-O', expanded)

        expanded = Phonemizer._expand_acronym('HaLO')
        self.assertEqual('Ha-L-O', expanded)

        expanded = Phonemizer._expand_acronym('vCPU')
        self.assertEqual('v-C-P-U', expanded)



import os
import shutil
import tempfile
import unittest
from pathlib import Path

import numpy as np
import torch

from dp import preprocess
from dp.model.model import AutoregressiveTransformer
from dp.model.predictor import Predictor
from dp.preprocess import preprocess
from dp.train import train
from dp.utils.io import read_config, save_config


class TestAutoregTraining(unittest.TestCase):

    def setUp(self) -> None:
        torch.manual_seed(42)
        np.random.seed(42)
        temp_dir = tempfile.mkdtemp(prefix='TestPreprocessTmp')
        self.temp_dir = Path(temp_dir)
        test_path = os.path.dirname(os.path.abspath(__file__))
        self.test_config_path = Path(test_path) / 'resources/autoreg_test_config.yaml'

    def tearDown(self) -> None:
        shutil.rmtree(self.temp_dir, ignore_errors=True)

    def test_autoregtraining_happy_path(self) -> None:
        config = read_config(self.test_config_path)
        data_dir = self.temp_dir / 'datasets'
        checkpoint_dir = self.temp_dir / 'checkpoints'
        config_path = self.temp_dir / 'autoreg_test_config.yaml'
        config['paths']['data_dir'] = str(data_dir)
        config['paths']['checkpoint_dir'] = str(checkpoint_dir)
        save_config(config, config_path)

        train_data = [
            ('en_us', 'young', 'jʌŋ'),
            ('de', 'benützten', 'bənʏt͡stn̩'),
            ('de', 'gewürz', 'ɡəvʏʁt͡s')
        ] * 100

        val_data = [
            ('en_us', 'young', 'jʌŋ'),
            ('de', 'gewürz', 'ɡəvʏʁt͡s')
        ] * 10

        preprocess(config_file=config_path,
                   train_data=train_data,
                   val_data=val_data,
                   deduplicate_train_data=False)

        train(rank=0, num_gpus=0, config_file=config_path)

        predictor = Predictor.from_checkpoint(checkpoint_dir / 'latest_model.pt')

        self.assertTrue(isinstance(predictor.model, AutoregressiveTransformer))

        result = predictor(words=['young'], lang='en_us')[0]
        self.assertEqual('jʌŋ', result.phonemes)
        self.assertTrue(result.confidence > 0.98)

        result = predictor(words=['gewürz'], lang='de')[0]
        self.assertEqual('ɡəvʏʁt͡s', result.phonemes)
        self.assertTrue(result.confidence > 0.96)

        result = predictor(words=['gewürz'], lang='en_us')[0]
        self.assertTrue(0.8 < result.confidence < 0.84)

        # test jit export
        predictor.model = torch.jit.script(predictor.model)
        result_jit = predictor(words=['gewürz'], lang='en_us')[0]
        self.assertEqual(result.phonemes, result_jit.phonemes)
        self.assertEqual(result.confidence, result_jit.confidence)



import unittest
import torch
from dp.training.dataset import PhonemizerDataset, collate_dataset


class TestPhonemizerDataset(unittest.TestCase):

    def test_get_items(self) -> None:
        items_input = [
            (0, [1, 2], [3, 4]),
            (1, [2, 3], [4, 5, 6])
        ]
        items = PhonemizerDataset(items=items_input)

        self.assertEqual(0, items[0]['item_id'])
        self.assertEqual(0, items[0]['language'])
        self.assertEqual([1, 2], items[0]['text'].tolist())
        self.assertEqual([3, 4], items[0]['phonemes'].tolist())
        self.assertEqual(2, items[0]['text_len'])
        self.assertEqual(2, items[0]['phonemes_len'])
        self.assertEqual(3, items[0]['start_index'])

        self.assertEqual(1, items[1]['item_id'])
        self.assertEqual(1, items[1]['language'])

    def test_collate(self):
        batch = [
            {
                'item_id': 0, 'language': 0, 'text_len': 2,
                'phonemes_len': 2, 'start_index': 2,
                'text': torch.tensor([1, 2]),
                'phonemes': torch.tensor([2, 3])
            },
            {
                'item_id': 1, 'language': 1, 'text_len': 3,
                'phonemes_len': 3, 'start_index': 3,
                'text': torch.tensor([1, 2, 3]),
                'phonemes': torch.tensor([3, 4, 5])
            }
        ]

        batch = collate_dataset(batch)
        self.assertEqual(0, batch['item_id'][0])
        self.assertEqual(1, batch['item_id'][1])
        self.assertEqual(2, batch['text_len'][0])
        self.assertEqual(3, batch['text_len'][1])
        self.assertEqual(2, batch['phonemes_len'][0])
        self.assertEqual(3, batch['phonemes_len'][1])
        self.assertEqual(0, batch['language'][0])
        self.assertEqual(1, batch['language'][1])
        self.assertEqual([1, 2, 0], batch['text'][0].tolist())
        self.assertEqual([1, 2, 3], batch['text'][1].tolist())
        self.assertEqual([2, 3, 0], batch['phonemes'][0].tolist())
        self.assertEqual([3, 4, 5], batch['phonemes'][1].tolist())



import unittest

from dp.training.metrics import word_error, phoneme_error


class TestWordError(unittest.TestCase):

    def test_call(self):
        predicted = ['a', 'b', 'c', 'd']
        target = ['a', 'k', 'c', 'a']
        result = word_error(predicted, target)
        self.assertEqual(1, result)

        predicted = ['r', 'r', 'r', 'r']
        target = ['a', 'k', 'c', 'a']
        result = word_error(predicted, target)
        self.assertEqual(1., result)

        predicted = ['a']
        target = ['a']
        result = word_error(predicted, target)
        self.assertEqual(0, result)


class TestPhonemeErrorRate(unittest.TestCase):

    def test_call(self):
        predicted = ['a', 'b', 'c', 'd']
        target = ['a', 'k', 'c', 'a']
        e, c = phoneme_error(predicted, target)
        self.assertEqual(2, e)
        self.assertEqual(4, c)

        predicted = ['r', 'r', 'r', 'r']
        target = ['a', 'k', 'c', 'a']
        e, c = phoneme_error(predicted, target)
        self.assertEqual(4, e)
        self.assertEqual(4, c)

        predicted = ['a']
        target = ['a']
        e, c = phoneme_error(predicted, target)
        self.assertEqual(0, e)
        self.assertEqual(1, c)

        predicted = ['a']
        target = ['b']
        e, c = phoneme_error(predicted, target)
        self.assertEqual(1, e)
        self.assertEqual(1, c)

        predicted = ['a', 'b']
        target = ['b']
        e, c = phoneme_error(predicted, target)
        self.assertEqual(1, e)
        self.assertEqual(1, c)

        predicted = ['a', 'b', 'c']
        target = ['b']
        e, c = phoneme_error(predicted, target)
        self.assertEqual(2, e)
        self.assertEqual(1, c)

        predicted = ['a']
        target = ['a', 'b']
        e, c = phoneme_error(predicted, target)
        self.assertEqual(1, e)
        self.assertEqual(2, c)



from typing import Dict, List, Tuple

import torch
from torch.nn.utils.rnn import pad_sequence

from dp import Prediction
from dp.model.model import load_checkpoint
from dp.model.utils import _get_len_util_stop
from dp.preprocessing.text import Preprocessor
from dp.preprocessing.utils import _batchify, _product


class Predictor:

    """ Performs model predictions on a batch of inputs. """

    def __init__(self,
                 model: torch.nn.Module,
                 preprocessor: Preprocessor) -> None:
        """
        Initializes a Predictor object with a trained transformer model a preprocessor.

        Args:
            model (Model): Trained transformer model.
            preprocessor (Preprocessor): Preprocessor corresponding to the model configuration.
        """

        self.model = model
        self.text_tokenizer = preprocessor.text_tokenizer
        self.phoneme_tokenizer = preprocessor.phoneme_tokenizer

    def __call__(self,
                 words: List[str],
                 lang: str,
                 batch_size: int = 8,
                 num_prons: int = 1) -> List[Prediction]:
        """
        Predicts phonemes for a list of words.

        Args:
          words (list): List of words to predict.
          lang (str): Language of texts.
          batch_size (int): Size of batch for model input to speed up inference.
          num_prons (int): Number of pronunciations to predict for each word. (Default value = 1)

        Returns:
          num_prons <= 1: List[Prediction]: A list of result objects containing (word, phonemes, phoneme_tokens, token_probs, confidence)
          num_prons >= 2: List[List[Prediction]]
        """

        predictions = dict()
        valid_texts = set()

        # handle words that result in an empty input to the model
        for word in words:
            input = self.text_tokenizer(sentence=word, language=lang)
            decoded = self.text_tokenizer.decode(
                sequence=input, remove_special_tokens=True)
            if len(decoded) == 0:
                if num_prons > 1:
                    predictions[word] = [([], [])]
                else:
                    predictions[word] = ([], [])
            else:
                valid_texts.add(word)

        valid_texts = sorted(list(valid_texts), key=lambda x: len(x))
        batch_pred = self._predict_batch(texts=valid_texts, batch_size=batch_size,
                                         language=lang,
                                         num_prons=num_prons)
        predictions.update(batch_pred)

        output = []
        for word in words:
            if num_prons > 1:
                pronunciations = []
                for (tokens, probs) in predictions[word]:
                    out_phons = self.phoneme_tokenizer.decode(sequence=tokens, remove_special_tokens=True)
                    out_phons_tokens = self.phoneme_tokenizer.decode(sequence=tokens, remove_special_tokens=False)
                    pronunciations.append(Prediction(word=word,
                                          phonemes=''.join(out_phons),
                                          phoneme_tokens=out_phons_tokens,
                                          confidence=_product(probs),
                                          token_probs=probs))
                output.append(pronunciations)
            else:
                tokens, probs = predictions[word]
                out_phons = self.phoneme_tokenizer.decode(
                    sequence=tokens, remove_special_tokens=True)
                out_phons_tokens = self.phoneme_tokenizer.decode(
                    sequence=tokens, remove_special_tokens=False)
                output.append(Prediction(word=word,
                                        phonemes=''.join(out_phons),
                                        phoneme_tokens=out_phons_tokens,
                                        confidence=_product(probs),
                                        token_probs=probs))

        return output

    def _predict_batch(self,
                       texts: List[str],
                       batch_size: int,
                       language: str,
                       num_prons: int) \
            -> Dict[str, Tuple[List[int], List[float]]]:
        """
        Returns dictionary with key = word and val = Tuple of (phoneme tokens, phoneme probs)
        """

        predictions = dict()
        text_batches = _batchify(texts, batch_size)
        for text_batch in text_batches:
            input_batch, lens_batch = [], []
            for text in text_batch:
                input = self.text_tokenizer(text, language)
                input_batch.append(torch.tensor(input))
                lens_batch.append(torch.tensor(len(input)))

            input_batch = pad_sequence(sequences=input_batch,
                                       batch_first=True, padding_value=0)
            lens_batch = torch.stack(lens_batch)
            start_indx = self.phoneme_tokenizer._get_start_index(language)
            start_inds = torch.tensor([start_indx]*input_batch.size(0)).to(input_batch.device)
            batch = {
                'text': input_batch,
                'text_len': lens_batch,
                'start_index': start_inds
            }
            with torch.no_grad():
                output_batch, probs_batch = self.model.generate(batch, num_prons=num_prons)

            if num_prons > 1:
                for text, output, probs in zip(text_batch, output_batch, probs_batch):
                    pronunciations = []
                    for _output, _probs in zip(output, probs):
                        _output, _probs = _output.cpu(), _probs.cpu()
                        seq_len = _get_len_util_stop(_output, self.phoneme_tokenizer.end_index)
                        pronunciations.append((_output[:seq_len].tolist(), _probs[:seq_len].tolist()))
                    predictions[text] = pronunciations
            else:
                output_batch, probs_batch = output_batch.cpu(), probs_batch.cpu()
                for text, output, probs in zip(text_batch, output_batch, probs_batch):
                    seq_len = _get_len_util_stop(output, self.phoneme_tokenizer.end_index)
                    predictions[text] = (output[:seq_len].tolist(), probs[:seq_len].tolist())

        return predictions

    @classmethod
    def from_checkpoint(cls, checkpoint_path: str, device='cpu') -> 'Predictor':
        """Initializes the predictor from a checkpoint (.pt file).

        Args:
          checkpoint_path (str): Path to the checkpoint file (.pt).
          device (str): Device to load the model on ('cpu' or 'cuda'). (Default value = 'cpu').

        Returns:
          Predictor: Predictor object.

        """
        model, checkpoint = load_checkpoint(checkpoint_path, device=device)
        preprocessor = checkpoint['preprocessor']
        return Predictor(model=model, preprocessor=preprocessor)




import unittest

from dp.training.evaluation import evaluate_samples


class TestEvaluation(unittest.TestCase):

    def test_evaluate_samples(self):

        lang_samples = {
            'de': [
                (['a'], ['a', 'b'], ['a', 'b']),
            ],
            'en_us': [
                # two alternative targets, evaluation should take the closer one with per=1./3
                (['a'], ['a', 'b', 'c'], ['b', 'b', 'b']),
                (['a'], ['a', 'b', 'c'], ['a', 'b', 'b'])
            ]
        }

        result = evaluate_samples(lang_samples=lang_samples)

        self.assertEqual(0., result['de']['per'])
        self.assertEqual(0., result['de']['wer'])
        self.assertEqual(1./3, result['en_us']['per'])
        self.assertEqual(1., result['en_us']['wer'])
        self.assertEqual((0 + 1) / (2 + 3), result['mean_per'])
        self.assertEqual((0 + 1) / 2, result['mean_wer'])


import unittest

from dp.preprocessing.text import SequenceTokenizer


class TestSequenceTokenizer(unittest.TestCase):

    def test_call_happy_path(self) -> None:
        symbols = ['a', 'b', 'c', 'd', 'e']
        languages = ['de', 'en']

        tokenizer = SequenceTokenizer(symbols=symbols, languages=languages, char_repeats=1,
                                      lowercase=True, append_start_end=True, end_token='<end>')

        tokens = tokenizer(['a', 'b'], language='de')
        decoded = tokenizer.decode(tokens)

        self.assertEqual(4, len(tokens))
        self.assertEqual(['<de>', 'a', 'b', '<end>'], decoded)

        tokens = tokenizer(['a', 'b'], language='en')
        decoded = tokenizer.decode(tokens)

        self.assertEqual(4, len(tokens))
        self.assertEqual(['<en>', 'a', 'b', '<end>'], decoded)

    def test_call_missing_symbols(self) -> None:
        symbols = ['a', 'b']
        languages = ['de', 'en']

        tokenizer = SequenceTokenizer(symbols=symbols, languages=languages,  lowercase=False,
                                      append_start_end=True, end_token='<end>', char_repeats=2)

        tokens = tokenizer(['A', 'b', 'F'], language='en')
        decoded = tokenizer.decode(tokens)

        self.assertEqual(4, len(tokens))
        self.assertEqual(['<en>', 'b', '<end>'], decoded)

    def test_call_edge_cases(self) -> None:
        symbols = ['a', 'b', 'c', 'd', 'e']
        languages = ['de', 'en']

        tokenizer = SequenceTokenizer(symbols=symbols, languages=languages,  lowercase=True,
                                      append_start_end=True, end_token='<end>', char_repeats=2)

        tokens = tokenizer([], language='en')
        decoded = tokenizer.decode(tokens)

        self.assertEqual(2, len(tokens))
        self.assertEqual(['<en>', '<end>'], decoded)

        tokens = tokenizer(['z'], language='en')
        decoded = tokenizer.decode(tokens)

        self.assertEqual(2, len(tokens))
        self.assertEqual(['<en>', '<end>'], decoded)

        tokenizer = SequenceTokenizer(symbols=['a'], languages=['de'],  lowercase=True,
                                      append_start_end=False, end_token=None, char_repeats=3)

        tokens = tokenizer(['a'], language='de')
        decoded = tokenizer.decode(tokens)

        self.assertEqual(3, len(tokens))
        self.assertEqual(['a'], decoded)

    def test_exception(self) -> None:
        tokenizer = SequenceTokenizer(symbols=['a'], languages=['de'],  lowercase=True,
                                      append_start_end=True, end_token='<end>', char_repeats=2)

        with self.assertRaises(ValueError):
            tokenizer(['a', 'b'], language='not_existent')

